/* die roller */
/* syntax chart:
 * dN		n-sided dice
 * oN		omit N lowest
 * ON		omit N highest
 * rN		reroll N or lower
 * RN		reroll N or higher
 * xN		roll N times
 * *N		==xN
 * +N		add N to total
 * -N		subtract N from total
 * t		totals
 * T		dice and totals
 * a		average only
 * A		average also
 * H		histogram
 * h		histogram only
 * f		fake numbers (histogram/average only)
 * N		N dice (default is d1)
 */

/* TODO:
 * * note that "-1" is wider than "9", so width is harder
 * * "cumulative" (+/-) flag for histograms
 */

#include <ctype.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

struct roll {
	int	nrolls;			/* number of distinct die rolls */
	int	*dice;			/* # of dice */
	int	*sides;			/* # sides of dice */
	int	*omit_l, *omit_h;	/* omit lowest, highest, N dice */
	int	*reroll_l, *reroll_h;	/* reroll l or lower, h or higher */
	int	mod;			/* modifier */
	int	times;			/* # of times to roll */
	int	flags;		/* see rollflags */
};

enum rollflags {
	DISPLAY_NONE,
	AUTODISPLAY_ROLLS,
	DISPLAY_ROLLS = (AUTODISPLAY_ROLLS << 1),
	DISPLAY_TOTAL = (DISPLAY_ROLLS << 1),
	DISPLAY_AVERAGE = (DISPLAY_TOTAL << 1),
	DISPLAY_HISTOGRAM = (DISPLAY_AVERAGE << 1),
	DISPLAY_ALL = ((DISPLAY_HISTOGRAM << 1) - 1),
	FAKE_HISTOGRAM = (DISPLAY_HISTOGRAM << 1)
};

#define O(x) offsetof(struct roll, x)

struct n2v {
	char *name, *descr, *letters;
	int def;
	size_t offset;
	enum { FALSE, TRUE } specify_many;
} name2value[] = {
	{ "dice", "number of dice", "n", 1, O(dice), TRUE },
	{ "sides", "sides on each die", "d", 1, O(sides), TRUE },
	{ "times", "times to roll", "x*", 1, O(times), FALSE },
	{ "omit_l", "omit N lowest dice", "o", 0, O(omit_l), TRUE },
	{ "omit_h", "omit N highest dice", "O", 0, O(omit_h), TRUE },
	{ "reroll_l", "reroll N or lower", "r", 0, O(reroll_l), TRUE },
	{ "reroll_h", "reroll N or higher", "R", 0, O(reroll_h), TRUE },
	{ 0, 0, 0, 0 },
};

struct n2f {
	char *name, *descr, *letters;
	int value;
} name2flag[] = {
	{ "rolls", "display each die", "v", DISPLAY_ROLLS },
	{ "total", "display total of each roll", "t", DISPLAY_TOTAL },
	{ "average", "display average of all rolls", "a", DISPLAY_AVERAGE },
	{ "histogram", "display histogram of all rolls", "h", DISPLAY_HISTOGRAM },
	{ "fake", "fake results", "f", FAKE_HISTOGRAM },
	{ 0, 0, 0, 0 },
};

struct macro {
	char *name, *value;
} *macros;
int nmacros = 0;

void printroll(struct roll *);
void doroll(struct roll *);
void freeroll(struct roll *);
int alloc(struct roll *, int);
int rollmax(struct roll *);
int rollmin(struct roll *);
int scanroll(struct roll *, char *);
int die(int);
int rnd(int);
int min(int *, int);
int max(int *, int);
int zmin(int *, int);
int zmax(int *, int);
int amax(int *, int);
int newmac(char *, char *);
int cpymac(char *, char *);
void showmac(void);
void delmac(char *);
void loadmac(char *);
void savemac(char *);
int dflag(struct roll *, int);
void omit_n(int *, int, int, int (*)(int *, int));

int
rnd(int n) {
	return n < 2 ? 0 : (random() % n);
}

int
die(int n) {
	return (n < 2 ? 0 : (random() % n)) + 1;
}

int
zmin(int *a, int n) {
	int m;
	int i;

	if (!a)
		return 0;
	/* make sure we don't start with a zero */
	m = max(a, n);
	for (i = 0; i < n; ++i) {
		if (a[i] && a[i] < m)
			m = a[i];
	}
	return m;
}

int
zmax(int *a, int n) {
	int m;
	int i;

	if (!a)
		return 0;
	m = a[0];
	for (i = 0; i < n; ++i) {
		if (a[i] && a[i] > m)
			m = a[i];
	}
	return m;
}

int
min(int *a, int n) {
	int m;
	int i;

	if (!a)
		return 0;
	m = a[0];
	for (i = 0; i < n; ++i) {
		if (a[i] < m)
			m = a[i];
	}
	return m;
}

int
amax(int *a, int n) {
	int m;
	int i;

	if (!a)
		return 0;
	m = abs(a[0]);
	for (i = 0; i < n; ++i) {
		if (abs(a[i]) > m)
			m = abs(a[i]);
	}
	return m;
}

int
max(int *a, int n) {
	int m;
	int i;

	if (!a)
		return 0;
	m = a[0];
	for (i = 0; i < n; ++i) {
		if (a[i] > m)
			m = a[i];
	}
	return m;
}

void
omit_n(int *a, int n, int omit, int (*minmax)(int *, int)) {
	int i, j, m;
	for (i = 0; i < omit; ++i) {
		m = (*minmax)(a, n);
		for (j = 0; j < n; ++j) {
			if (a[j] == m) {
				a[j] = 0;
				break;
			}
		}

	}
}
void
printroll(struct roll *r) {
	struct n2v *p;
	struct n2f *q;
	int i, *ip;

	if (!r) {
		printf("null-roll");
		return;
	}

	for (i = 0; i < r->nrolls; ++i) {
		for (p = name2value; p->name; ++p) {
			if (!p->specify_many)
				continue;
			ip = *(int **) ((char *)r + p->offset);
			if (ip && ip[i] && ip[i] != p->def) {
				if (p->letters[0] != 'n')
					printf("%c%d", p->letters[0], ip[i]);
				else
					printf("%d", ip[i]);
			}
		}
		if (i < (r->nrolls - 1)) {
			if (r->dice[i + 1] > 0) {
				putchar('+');
			} else {
				putchar('-');
			}
		}
	}
	if (r->mod) {
		printf("%+d", r->mod);
	}
	for (p = name2value; p->name; ++p) {
		if (p->specify_many)
			continue;
		ip = (int *) ((char *)r + p->offset);
		if (*ip && *ip != p->def) {
			if (p->letters[0] != 'n')
				printf("%c%d", p->letters[0], *ip);
			else
				printf("%d", *ip);
		}
	}

	if (dflag(r, DISPLAY_ALL) != DISPLAY_ROLLS) {
		for (q = name2flag; q->name; ++q) {
			if (q->letters[0] && r->flags & q->value) {
				printf("%c", q->letters[0]);
			}
		}
	}
}

void
doroll(struct roll *r) {
	long double avg = 0;
	int *histogram = 0;
	int *dice;
	int hcount = 0;
	int dwidth, rwidth, dcount = 0;
	int dice_per_line;
	int total;
	int roll_number, sub;
	int rmin, rmax;
	int d;
	int i, j, k, n;
	int linepos;
	int negative;
	int maxdice;
	char dummy;

	if (!r)
		return;

	rmin = rollmin(r);
	rmax = rollmax(r);

	if (dflag(r, DISPLAY_HISTOGRAM)) {
		 hcount = rmax - rmin + 1;
		 histogram = malloc(hcount * sizeof(*histogram));
		 if (!histogram) {
			fprintf(stderr, "couldn't allocate space for histogram.\n");
			r->flags = (r->flags & ~DISPLAY_ALL) |
				(r->flags & DISPLAY_ALL & ~DISPLAY_HISTOGRAM);
		 }
		 for (i = 0; i < hcount; ++i)
			histogram[i] = 0;
	}
	if (r->flags & FAKE_HISTOGRAM)
		r->times = 0;

	dwidth = snprintf(&dummy, 1, "%d", max(r->sides, r->nrolls));
	rwidth = snprintf(&dummy, 1, "%d", rmax);
	/* -9 is wider than 8 */
	i = snprintf(&dummy, 1, "%d", rmin);
	if (i > rwidth)
		rwidth = i;

	dice_per_line = 78 / (dwidth + 1);

	/* absolute value max */
	maxdice = amax(r->dice, r->nrolls);
	dice = malloc(maxdice * sizeof(int));
	if (!dice) {
		fprintf(stderr, "help!  can't allocate dice\n");
		if (dflag(r, DISPLAY_HISTOGRAM))
			free(histogram);
		return;
	}

	for (roll_number = 0; roll_number < r->times; ++roll_number) {
		total = r->mod;
		for (sub = 0; sub < r->nrolls; ++sub) {
			if (r->dice[sub] < 0) {
				negative = 1;
				r->dice[sub] *= -1;
			} else {
				negative = 0;
			}

			n = r->sides[sub] - r->reroll_l[sub];
			if (r->reroll_h[sub])
				n -= (r->sides[sub] - r->reroll_h[sub]);

			for (d = 0; d < r->dice[sub]; ++d) {
				dice[d] = die(n) + r->reroll_l[sub];
			}

			/* drop lowest and highest */
			omit_n(dice, r->dice[sub], r->omit_l[sub], zmin);
			omit_n(dice, r->dice[sub], r->omit_h[sub], zmax);

			if (dflag(r, DISPLAY_ROLLS)) {
				for (d = 0; d < r->dice[sub]; ++d) {
					if (!dice[d])
						continue;
					/* XXX should count dice printed */
					printf("%-*d%c", dwidth, dice[d],
						((dcount + 1) % dice_per_line) ? ' ' : '\n');
					++dcount;
				}
			}

			if (negative) {
				for (d = 0; d < r->dice[sub]; ++d) {
					total -= dice[d];
				}
				r->dice[sub] *= -1;
				negative = 0;
			} else {
				for (d = 0; d < r->dice[sub]; ++d) {
					total += dice[d];
				}
			}
		}
		linepos = (dcount % dice_per_line) * dwidth;
		if (dflag(r, DISPLAY_ROLLS) && r->mod) {
			if (snprintf(&dummy, 1, "%+d", r->mod) + linepos > 78 - rwidth) {
				putchar('\n');
				linepos = 0;
			}
			linepos += printf("%+d", r->mod);
		}
		if (dflag(r, DISPLAY_TOTAL) && dflag(r, DISPLAY_ROLLS)) {
			printf(": ");
		}
		if (dflag(r, DISPLAY_TOTAL)) {
			printf("%d", total);
		}
		if (dflag(r, DISPLAY_HISTOGRAM))
			histogram[total - rmin]++;
		if (dflag(r, DISPLAY_AVERAGE))
			avg += total;
		if (dflag(r, DISPLAY_TOTAL) || dflag(r, DISPLAY_ROLLS))
			putchar('\n');
	}
	if (r->flags & FAKE_HISTOGRAM) {
		int *results, *scratchstack, *minstack, *maxstack;
		int *rollstack;
		int totaldice = 0;
		struct { int low, high, dice, start; } *omits, *op;
		int totalomits = 0;
		int times = 0;
		total = 0;
		avg = 0;

		for (i = 0; i < r->nrolls; ++i) {
			if (r->dice[i] > 0) {
				totaldice += r->dice[i];
			} else {
				totaldice -= r->dice[i];
			}
			if (r->omit_l[i] || r->omit_h[i]) {
				totalomits++;
			}
		}
		results = malloc(totaldice * sizeof(int));
		if (!results) {
			fprintf(stderr, "no space for faked results\n");
			goto quit;
		}
		rollstack = malloc(totaldice * sizeof(int));
		if (!rollstack) {
			free(results);
			fprintf(stderr, "no space for roll indicies for faked results\n");
			goto quit;
		}
		minstack = malloc(totaldice * sizeof(int));
		if (!minstack) {
			free(results);
			free(rollstack);
			fprintf(stderr, "no space for roll minima for faked results\n");
			goto quit;
		}
		maxstack = malloc(totaldice * sizeof(int));
		if (!maxstack) {
			free(results);
			free(rollstack);
			free(minstack);
			fprintf(stderr, "no space for roll maxima for faked results\n");
			goto quit;
		}
		if (totalomits) {
			scratchstack = malloc(totaldice * sizeof(int));
			if (!scratchstack) {
				free(results);
				free(rollstack);
				free(minstack);
				free(maxstack);
				fprintf(stderr, "no space for roll maxima for faked results\n");
				goto quit;
			}
			omits = malloc(totalomits * sizeof(*omits));
			if (!omits) {
				free(results);
				free(rollstack);
				free(minstack);
				free(maxstack);
				free(scratchstack);
				fprintf(stderr, "no space for roll maxima for faked results\n");
				goto quit;
			}
		}
		/* build list of things to integrate */
		k = 0;
		op = omits;
		for (i = 0; i < r->nrolls; ++i) {
			if (r->dice[i] < 0) {
				negative = 1;
				r->dice[i] *= -1;
			} else {
				negative = 0;
			}
			if (r->omit_l[i] || r->omit_h[i]) {
				op->low = r->omit_l[i];
				op->high = r->omit_h[i];
				op->start = k;
				op->dice = r->dice[i];
				++op;
			}
			for (j = 0; j < r->dice[i]; ++j) {
				minstack[k] = r->reroll_l[i] + 1;
				maxstack[k] = r->reroll_h[i] ?
					r->reroll_h[i] - 1 :
					r->sides[i];
				results[k] = minstack[k];
				rollstack[k] = i;
				++k;
			}
			if (negative) {
				r->dice[i] *= -1;
			}
		}
		while (results[totaldice - 1] <= maxstack[totaldice - 1]) {
			++times;
			total = r->mod;
			if (totalomits) {
				memcpy(scratchstack, results, totaldice * sizeof(int));
			}
			for (i = 0; i < totalomits; ++i) {
				int *ip = results + omits[i].start;
				omit_n(ip, omits[i].dice, omits[i].low, zmin);
				omit_n(ip, omits[i].dice, omits[i].high, zmax);
			}
			for (i = 0; i < totaldice; ++i) {
				if (r->dice[rollstack[i]] > 0) {
					total += results[i];
				} else {
					total -= results[i];
				}
			}
			if (totalomits) {
				memcpy(results, scratchstack, totaldice * sizeof(int));
			}
			if (r->flags & DISPLAY_HISTOGRAM) {
				++histogram[total - rmin];
			}
			if (r->flags & DISPLAY_AVERAGE) {
				avg += total;
			}
			++results[0];
			k = 0;
			while ((results[k] > maxstack[k]) &&
				(k < (totaldice - 1))) {
				results[k] = minstack[k];
				++k;
				results[k]++;
			}
		}
		r->times = times;
		free(results);
		free(rollstack);
		free(minstack);
		free(maxstack);
		if (totalomits) {
			free(scratchstack);
			free(omits);
		}
	}
quit:
	free(dice);
	if (dflag(r, DISPLAY_HISTOGRAM)) {
		int i, j, value;
		int hmin, hmax;
		int w;
		char dummy;
		double star;
		int hwidth;

		value = rmin;
		hmin = min(histogram, hcount);
		hmax = max(histogram, hcount);
		hwidth = snprintf(&dummy, 1, "%d", hmax);

		w = rwidth + 3 + hwidth;

		star = (hmax + 0.0) / (78 - w);

		for (i = 0, value = rmin; i < hcount; ++i, ++value) {
			printf("%*d: %*d: ",
				rwidth, value, hwidth, histogram[i]);
			for (j = 0; j < histogram[i] / star; ++j) {
				putchar('*');
			}
			putchar('\n');
		}
		free(histogram);
	}
	if (dflag(r, DISPLAY_AVERAGE)) {
		if (dflag(r, DISPLAY_ALL) == DISPLAY_AVERAGE) {
			printf("%Lf\n", avg / r->times);
		} else {
			printf("average: %Lf\n", avg / r->times);
		}
	}
}

int
rollmin(struct roll *r) {
	int total = 0;
	int count;
	int min;
	int i;
	if (!r)
		return 0;
	for (i = 0; i < r->nrolls; ++i) {
		if (r->dice[i] < 0) {
			count = -r->dice[i] - (r->omit_l[i] + r->omit_h[i]);
			if (r->reroll_h[i]) {
				min = r->reroll_h[i] - 1;
			} else {
				min = r->sides[i];
			}
			total -= count * min;
		} else {
			count = r->dice[i] - (r->omit_l[i] + r->omit_h[i]);
			if (r->reroll_l[i]) {
				min = r->reroll_l[i] + 1;
			} else {
				min = 1;
			}
			total += count * min;
		}
	}
	return total + r->mod;
}

int
rollmax(struct roll *r) {
	int total = 0;
	int count;
	int max;
	int i;
	if (!r)
		return 0;
	for (i = 0; i < r->nrolls; ++i) {
		if (r->dice[i] < 0) {
			count = -r->dice[i] - (r->omit_l[i] + r->omit_h[i]);
			if (r->reroll_l[i]) {
				max = r->reroll_l[i] + 1;
			} else {
				max = 1;
			}
			total -= count * max;
		} else {
			count = r->dice[i] - (r->omit_l[i] + r->omit_h[i]);
			if (r->reroll_h[i]) {
				max = r->reroll_h[i] - 1;
			} else {
				max = r->sides[i];
			}
			total += count * max;
		}
	}
	return total + r->mod;
}

int
alloc(struct roll *r, int n) {
	int *ip, **vp;
	if (!r)
		return 0;
	if (n < 1) {
		freeroll(r);
		return 0;
	}
	if (n > r->nrolls) {
		struct n2v *p;
		for (p = name2value; p->name; ++p) {
			if (!p->specify_many)
				continue;
			vp = (int **) ((char *) r + p->offset);
			ip = malloc(n * sizeof(int));
			if (ip) {
				if (*vp) {
					memcpy(ip, *vp, r->nrolls * sizeof(int));
					free(*vp);
				}
				*vp = ip;
			} else {
				freeroll(r);
				return 0;
			}
		}
	}
	r->nrolls = n;
	return 1;
}

void
freeroll(struct roll *r) {
	struct n2v *p;
	int **vp;

	if (!r)
		return;

	for (p = name2value; p->name; ++p) {
		if (!p->specify_many)
			continue;
		vp = (int **) ((char *) r + p->offset);
		if (*vp) {
			free(*vp);
			*vp = NULL;
			
		}
	}
	r->nrolls = 0;
}

int
dflag(struct roll *r, int flag) {
	return (r->flags & DISPLAY_ALL) & flag;
}

int
scanroll(struct roll *r, char *s) {
	char *t;
	long l;
	struct n2v *p;
	struct n2f *q;
	struct roll defaults = { 0, 0, 0, 0, 0, 0, 0, 0, 1, AUTODISPLAY_ROLLS };
	int negative = 0;

	*r = defaults;

	if (!alloc(r, 1))
		return 0;

	r->dice[0] = 1;
	r->sides[0] = 1;
	r->omit_h[0] = 0;
	r->omit_l[0] = 0;
	r->reroll_h[0] = 0;
	r->reroll_l[0] = 0;

	while (*s) {
		if (isdigit(*s)) {
			l = strtol(s, &t, 0);
			if (l >= 0) {
				r->dice[r->nrolls - 1] = l;
			} else {
				fprintf(stderr, "Couldn't figure out number.\n");
				return 1;
			}
			s = t;
			if (negative) {
				r->dice[r->nrolls - 1] *= -1;
			}
			goto done;
		}
		if (*s == '+' || *s == '-') {
			/* turn it into a +N */
			if (r->sides[r->nrolls - 1] == 1) {
				r->mod += r->dice[r->nrolls - 1];
			} else {
				if (!alloc(r, r->nrolls + 1))
					return 0;
			}
			if (*s == '-') {
				negative = 1;
			} else {
				negative = 0;
			}
			r->dice[r->nrolls - 1] = 1;
			r->sides[r->nrolls - 1] = 1;
			r->omit_h[r->nrolls - 1] = 0;
			r->omit_l[r->nrolls - 1] = 0;
			r->reroll_h[r->nrolls - 1] = 0;
			r->reroll_l[r->nrolls - 1] = 0;
			++s;
			goto done;
		}
		for (p = name2value; p->name; ++p) {
			if (strchr(p->letters, *s)) {
				l = strtol(s + 1, &t, 0);
				if (l >= 0) {
					if (p->specify_many) {
						int *ip = *((int **) ((char *) r + p->offset));
						ip[r->nrolls - 1] = l;
					} else {
						int i = l;

						memcpy((char *)r + p->offset,
							&i, sizeof(i));
					} 
				} else {
						fprintf(stderr, "Couldn't figure out number.\n");
						return 1;
					}
				s = t;
				goto done;
			}
		}
		for (q = name2flag; q->name; ++q) {
			if (strchr(q->letters, *s)) {
				r->flags |= q->value;
				++s;
				goto done;
			} else if (strchr(q->letters, tolower(*s))) {
				r->flags |= q->value;
				if (dflag(r, AUTODISPLAY_ROLLS)) {
					r->flags = (r->flags & ~DISPLAY_ALL) |
						(r->flags & DISPLAY_ALL & ~AUTODISPLAY_ROLLS);
					r->flags |= DISPLAY_ROLLS;
				}
				++s;
				goto done;
			}
		}
		if (*s) {
			fprintf(stderr, "Unknown character `%c'\n", *s);
			++s;
		}
done:		;
	}
	if (r->sides[r->nrolls - 1] == 1) {
		r->mod += r->dice[r->nrolls - 1];
		--r->nrolls;
	}
	if (dflag(r, DISPLAY_ALL) == AUTODISPLAY_ROLLS) {
		r->flags = DISPLAY_ROLLS | (r->flags & ~DISPLAY_ALL);
	}
	if ((r->nrolls > 1 || r->mod) && dflag(r, DISPLAY_ROLLS)) {
		r->flags |= DISPLAY_TOTAL;
	}
	return 1;
}

int
newmac(char *name, char *value) {
	struct macro *m;

	if (!name || !value || !strlen(name) || !strlen(value))
		return 0;

	for (m = macros; m && m->name; ++m) {
		if (!strcmp(m->name, name)) {
			free(m->value);
			m->value = strdup(value);
			return 1;
		}
	}

	m = realloc(macros, sizeof(struct macro) * (nmacros + 2));
	if (!m) {
		fprintf(stderr, "can't allocate space for new macro\n");
		return 0;
	} else {
		macros = m;
	}
	macros[nmacros].name = strdup(name);
	macros[nmacros].value = strdup(value);
	++nmacros;
	macros[nmacros].name = 0;
	macros[nmacros].value = 0;
	return 1;
}

int
cpymac(char *dest, char *src) {
	struct macro *m;
	for (m = macros; m && m->name; ++m) {
		if (!strcmp(m->name, src)) {
			strcpy(dest, m->value);
			return 1;
		}
	}
	return 0;
}

void
delmac(char *name) {
	struct macro *m, *n;
	n = macros;
	for (m = macros; m && m->name; ++m) {
		if (strcmp(m->name, name)) {
			*n++ = *m;
		} else {
			--nmacros;
			free(m->name);
			free(m->value);
		}
	}
	if (m)
		*n = *m;
}

void
showmac(void) {
	struct macro *m;
	for (m = macros; m && m->name; ++m) {
		printf("%s: %s\n", m->name, m->value);
	}
}

void
savemac(char *name) {
	FILE *f;
	struct macro *m;

	if (!name)
		return;
	f = fopen(name, "w");
	if (!f) {
		fprintf(stderr, "couldn't open file `%s'\n", name);
		return;
	}

	printf("saving macros to `%s': ", name);
	fflush(stdout);
	for (m = macros; m && m->name; ++m) {
		fprintf(f, "%s:%s\n", m->name, m->value);
		putchar('.');
		fflush(stdout);
	}
	putchar('\n');
}

void
loadmac(char *name) {
	char s[1024], *t;
	char *n, *v;
	int count = 0;
	FILE *f;

	if (!name)
		return;
	f = fopen(name, "r");
	if (!f) {
		fprintf(stderr, "couldn't open file `%s'\n", name);
		return;
	}
	printf("loading macros from `%s': ", name);
	fflush(stdout);
	while (fgets(s, 1024, f)) {
		++count;
		t = strchr(s, '\n');
		if (t)
			*t = '\0';
		n = s;
		v = s;
		strsep(&v, ":");
		if (n && v && strlen(n) && strlen(v)) {
			putchar('.');
			fflush(stdout);
			newmac(n, v);
		} else {
			fprintf(stderr, "error reading line %d of `%s'\n",
				count, name);
			break;
		}
	}
	putchar('\n');
	fclose(f);
}

int
main(int argc, char **argv) {
	struct roll r;
	int i;
	char s[1024];
	char previous[1024];
	char scratch[1024];
	char macroexpn[1024];
	char dicefile[1024];
	char *t, *u;

	t = getenv("HOME");
	if (t) {
		if (snprintf(dicefile, 1024, "%s/.dicerc", t) > 1023) {
			fprintf(stderr, "warning:  your home directory is too deep.\n");
		}
	} else {
		strcpy(dicefile, ".dicerc");
	}
	srandom(time(NULL) + getpid());
	if (argc > 1) {
		for (i = 1; i < argc; ++i) {
			if (scanroll(&r, argv[i])) {
				doroll(&r);
			}
		}
		exit(EXIT_SUCCESS);
	}
	while (printf("? "), fflush(stdout), fgets(s, 1024, stdin)) {
		t = strchr(s, '\n');
		if (t)
			*t = '\0';
	parse:
		if (strlen(s) < 2 || strchr(s, ' ')) {
			switch(*s) {
			case '\0':
				strcpy(s, previous);
				printf("<%s>\n", s);
				goto parse;
				break;
			case 's':
				showmac();
				break;
			case 'q':
				printf("bye-bye.\n");
				exit(EXIT_SUCCESS);
				break;
			case 'w':
				t = strrchr(s, ' ');
				if (!t)
					t = dicefile;
				else
					++t;
				savemac(t);
				break;
			case 'l':
				t = strrchr(s, ' ');
				if (!t)
					t = dicefile;
				else
					++t;
				loadmac(t);
				break;
			default:
				fprintf(stderr, "don't confuse me.\n");
				break;
			}
		} else {
			if ((t = strchr(s, '<')) && (u = strchr(t, '>'))) {
				*u = '\0';
				*t = '\0';
				if (cpymac(macroexpn, t + 1)) {
					sprintf(scratch, "%s%s%s",
						s, macroexpn, u + 1);
					strcpy(s, scratch);
					goto parse;
				} else {
					fprintf(stderr, "unknown macro `%s'\n", t + 1);
					continue;
				}
			}
			switch (*s) {
			case '>':
				newmac(s + 1, previous);
				break;
			case '<':
				cpymac(s, s + 1);
				goto parse;
			case '^':
				if (s[1] == '^') {
					delmac(NULL);
				} else {
					delmac(s + 1);
				}
				break;
			case '!':
				system(s+1);
				break;
			default:
				if (strpbrk(s, "0123456789")) {
					if (scanroll(&r, s)) {
						strcpy(previous, s);
						doroll(&r);
					} else {
						fprintf(stderr, "can't figure out roll\n");
						continue;
					}
				} else {
					i = strspn(s, "abcdefghijklmnopqrstuvwxyz");
					if (i < strlen(s)) {
						sprintf(scratch, "%.*s", i, s);
						if (cpymac(macroexpn, scratch)) {
							sprintf(scratch, "%s%s",
								macroexpn,
								s + i);
							strcpy(s, scratch);
							goto parse;
						} else {
							fprintf(stderr,
								"unknown macro `%s'\n",
								scratch);
							continue;
						}
					} else if (cpymac(s, s)) {
						goto parse;
					} else {
						fprintf(stderr, "unknown command `%s'\n", s);
						continue;
					}
				}
			}
		}
	}
	return 0;
}
