#include "list.h"

/* equivalent, but the former is provably legal.  What are we doing?
 * If *x's first member is ptr-to-structure, then x can be treated as
 * ptr-to-ptr-to-structrue, with an appropriate cast, because a pointer
 * to a struct is a pointer to its first member.
 */
#ifdef __STDC__
#define NEXT(x) (*((ls_list **)(x)))
#else
#define NEXT(x) (((ls_list *)(x))->next)
#endif

static int (*lcmp)(const void *, const void *);
static void *local_lsort(void *, int);
static ls_list * merge(void *, void *);

ls_list *
ls_new(size_t sz, void (*f)(void *)) {
	ls_list *new;
	new = malloc(sz);
	if (new) {
		NEXT(new) = NULL;
		if (f)
			f(new);
	}
	return new;
}

ls_dlist *
ls_dnew(size_t sz, void (*f)(void *)) {
	ls_dlist *new;
	new = malloc(sz);
	if (new && f)
		f(new);
	return new;
}

void
ls_free(void *v, void (*f)(void *)) {
	ls_list *mp;

	while (v) {
		mp = NEXT(v);
		if (f)
			f(v);
		free(v);
		v = mp;
	}
}

void
ls_dfree(void *v, void (*f)(void *)) {
	ls_free(v, f);
}

ls_list *
ls_append(void *dest, void *src) {
	NEXT(src) = NEXT(dest);
	NEXT(dest) = src;
	return dest;
}

ls_list *
ls_push(void *dest, void *src) {
	NEXT(src) = dest;
	return src;
}

ls_list *
ls_iter(void *v, int (*f)(void *)) {
	while (v) {
		if (f && f(v))
			return v;
		v = NEXT(v);
	}
	return v;
}

ls_list *
ls_sort(void *source, int nitems, int (*cmp)(const void *, const void *)) {
	void *vp;
	int i;

	if (cmp)
		lcmp = cmp;

	if (!source)
		return 0;

	if (nitems == 0) {
		return 0;
	}

	if (nitems < 0) {
		i = 0;
		vp = source;
		while ((vp = NEXT(vp)) != 0)
			++i;
		nitems = i + 1;
	}

	if (nitems == 1)
		return source;

	vp = source;

	if (nitems == 2) {
		if (lcmp(vp, NEXT(vp)) > 0) {
			NEXT(NEXT(vp)) = vp;
			vp = NEXT(vp);
			NEXT(NEXT(vp)) = 0;
			return vp;
		} else {
			return vp;
		}
	}

	vp = local_lsort(source, nitems);

	return vp;
}

static void *
local_lsort(void *source, int nitems) {
	void *lp, *end, *mid;
	int i;

#ifdef COUNTS
	if (nitems < COUNT_MAX)
		++count_per[nitems];
#endif

	lp = source;

	/* special cases we do without merging */
	switch (nitems) {
	case 1:
		return source;
	case 2:
		/* if list is out of order... */
		if (lcmp(lp, NEXT(lp)) > 0) {
			end = NEXT(lp);
			NEXT(end) = lp;
			NEXT(lp) = 0;
			return end;
		} else {
			return lp;
		}
	case 3:
		mid = NEXT(lp);
		end = NEXT(mid);
		/* if head of list is after middle... */
		if (lcmp(lp, mid) > 0) {
			/* if head is also after end... */
			if (lcmp(lp, end) > 0) {
				/* if middle is after end... */
				if (lcmp(mid, end) > 0) { /* EMH */
					NEXT(end) = mid;
					NEXT(mid) = lp;
					NEXT(lp) = 0;
					return end;
				} else { /* middle is before end */ /* MEH */
					NEXT(end) = lp;
					NEXT(lp) = 0;
					return mid;
				}
			} else { /* head is before end */ /* MHE */
				NEXT(lp) = end;
				NEXT(mid) = lp;
				return mid;
			}
		} else { /* head of list is before middle */
			/* if head is after end */
			if (lcmp(lp, end) > 0) { /* EHM */
				NEXT(mid) = 0;
				NEXT(end) = lp;
				return end;
			} else {
				if (lcmp(mid, end) > 0) { /* HEM */
					NEXT(mid) = 0;
					NEXT(end) = mid;
					NEXT(lp) = end;
					return lp;
				} else { /* HME */
					return lp;
				}
			}
		}
	}

	i = nitems / 2;

	while (--i)
		lp = NEXT(lp);

	i = nitems / 2;
	mid = NEXT(lp);
	NEXT(lp) = 0;

	return merge(local_lsort(source, i), local_lsort(mid, nitems - i));
}

static ls_list *
merge(void *lp1, void *lp2) {
	ls_list *out, *walk;

	if (lcmp(lp1, lp2) > 0) {
		out = lp2;
		lp2 = NEXT(lp2);
	} else {
		out = lp1;
		lp1 = NEXT(lp1);
	}

	walk = out;

	while (lp1 && lp2) {
		while (lp1 && lcmp(lp1, lp2) < 1) {
			NEXT(walk) = lp1;
			walk = lp1;
			lp1 = NEXT(lp1);
		}
		if (!lp1)
			break;
		while (lp2 && lcmp(lp2, lp1) < 0) {
			NEXT(walk) = lp2;
			walk = lp2;
			lp2 = NEXT(lp2);
		}
	}

	if (lp1) {
		NEXT(walk) = lp1;
	} else {
		NEXT(walk) = lp2;
	}
	/* and it's already NULL terminated. */

	return out;
}

ls_list *
ls_cat(ls_list *l1, ls_list *l2) {
	ls_list *tl;

	for (tl = l1; tl && tl->next; tl = tl->next)
		;

	if (tl)
		tl->next = l2;

	return l1;
}

ls_list *
ls_end(ls_list *l) {
	for (;l && l->next; l = l->next)
		;

	return l;
}

typedef struct {
	ls_list *head, *tail;
} ls_twine;

static ls_twine ls_rrev(ls_list *);

ls_list *
ls_rev(ls_list *l) {
	ls_twine t;

	t = ls_rrev(l);
	return t.head;
}

ls_twine
ls_rrev(ls_list *l) {
	ls_twine t = { 0, 0 };

	if (!l)
		return t;

	if (!l->next) {
		t.head = l;
		t.tail = l;
		return t;
	}

	t = ls_rrev(l->next);
	t.tail->next = l;
	l->next = 0;
	t.tail = l;

	return t;
}
